import { Player, system, world } from "@minecraft/server";
import { distance } from "lib/maths";
import { getAllValues, getEffectColor, getGrowthPercentage, getHealth, getItems, getPack, getToolIcon, getValidValues, time } from "lib/display";
import { occupations } from "lib/global";
import "commands/configuration";
import { FormCancelationReason, ModalFormData } from "@minecraft/server-ui";
import { specialBlocks } from "lib/blocks";
world.afterEvents.playerSpawn.subscribe(({ player }) => {
    const defaultSettings = {
        waio: true,
        developer: false,
        noLang: false,
        maxDistance: 5,
        maxItemsDisplay: 6,
    };
    for (const [key, value] of Object.entries(defaultSettings)) {
        if (player.getDynamicProperty(key) === undefined)
            player.setDynamicProperty(key, value);
    }
});
system.runInterval(() => {
    const players = world.getAllPlayers();
    for (const player of players) {
        if (!player.getDynamicProperty("waio") || !player.isValid)
            return;
        const entity = player.getEntitiesFromViewDirection()[0]?.entity;
        const block = player.getBlockFromViewDirection()?.block;
        const headLoc = player.getHeadLocation();
        const sneaking = player.isSneaking;
        const developer = player.getDynamicProperty("developer");
        const noLang = player.getDynamicProperty("noLang");
        const maxDistance = player.getDynamicProperty("maxDistance");
        let titleRaw = "";
        let underRaw = "";
        if (entity && distance(headLoc, entity.location, maxDistance)) {
            const isItem = entity.getComponent("item");
            const health = entity.getComponent("health");
            const variant = entity.getComponent("variant")?.value;
            const markVar = entity.getComponent("mark_variant")?.value;
            const skinId = entity.getComponent("skin_id")?.value;
            const tameable = entity.getComponent("tameable");
            const movement = entity.getComponent("movement")?.defaultValue;
            const effects = entity?.getEffects();
            const isVillager = entity.typeId == 'minecraft:villager_v2' && occupations[variant - 1];
            const isntInanimate = entity.matches({ excludeFamilies: ["inanimate"] });
            if (entity.matches({ families: ["waio:no_display"] }))
                return;
            if (!isItem) {
                const id = entity.typeId.split(":")[0];
                const pack = getPack(id, noLang);
                titleRaw += developer ? `\n§7${entity.typeId}` : "";
                if (health && isntInanimate) {
                    const { currentValue, defaultValue } = health;
                    titleRaw += `\n${getHealth(Math.round(currentValue), defaultValue)}`;
                }
                if (effects && effects.length > 0) {
                    const effectList = effects.map(({ displayName, duration, typeId }) => `\n${getEffectColor(typeId)}${displayName} (${time(duration)})`);
                    underRaw += effectList.join("");
                }
                if (developer && (variant || markVar || skinId)) {
                    if (variant)
                        underRaw += `\n§dVariant: ${variant}`;
                    if (markVar)
                        underRaw += `\n§dMark Variant: ${markVar}`;
                    if (skinId)
                        underRaw += `\n§dSkin Id: ${skinId}`;
                }
                if (developer && entity.getTags().length > 0) {
                    underRaw += `\n§3Tags: [${entity.getTags().join(", ")}]`;
                }
                player.onScreenDisplay.setActionBar({
                    rawtext: [
                        { translate: (entity instanceof Player || entity.nameTag) && !entity.matches({ families: ["waio:no_nametag"] }) ? entity.nameTag : entity.localizationKey },
                        { text: titleRaw },
                        { translate: isVillager ? "waio.entity.occupation" : "" },
                        { translate: isVillager ? `entity.villager.${occupations[variant - 1]}` : "" },
                        { translate: sneaking && movement && isntInanimate ? "waio.entity.speed" : "" },
                        { text: sneaking && movement && isntInanimate ? movement.toFixed(2) : "" },
                        { translate: sneaking && tameable ? "waio.entity.tame_chance" : "" },
                        { text: sneaking && tameable ? tameable.probability.toFixed(2) : "" },
                        { translate: sneaking && tameable ? "waio.entity.tame_items" : "" },
                        { text: sneaking && tameable ? tameable?.getTameItems.map((item) => item.typeId).join(', ') + '.' : "" },
                        { text: underRaw },
                        { translate: `\n§9§o${pack}§r` },
                    ]
                });
            }
            else {
                const item = isItem.itemStack;
                const id = item.typeId.split(":")[0];
                const pack = getPack(id, noLang);
                const entityTags = entity.getTags();
                const itemTags = item.getTags();
                titleRaw += developer ? `\n§7${item.typeId}` : "";
                if (developer && (entityTags.length > 0 || itemTags.length > 0)) {
                    if (itemTags.length > 0)
                        underRaw += `\n§3Item Tags: [${itemTags.join(", ")}]`;
                    if (entityTags.length > 0)
                        underRaw += `\n§3Entity Tags: [${entityTags.join(", ")}]`;
                }
                player.onScreenDisplay.setActionBar({
                    rawtext: [
                        { translate: item.nameTag ? item.nameTag : item.localizationKey },
                        { text: " x" + item.amount },
                        { text: titleRaw },
                        { text: underRaw },
                        { translate: `\n§9§o${pack}§r` },
                    ]
                });
            }
        }
        else if (block && distance(headLoc, block.location, maxDistance)) {
            const { typeId, permutation, location } = block;
            let [id, name] = typeId.split(":");
            const pack = getPack(id, noLang);
            const displayIcon = getToolIcon(block);
            const states = block.permutation.getAllStates();
            const maxItemsDisplay = player.getDynamicProperty("maxItemsDisplay");
            const inv = block.getComponent("inventory")?.container;
            titleRaw += developer ? `\n§7${typeId}` : "";
            titleRaw += `\n§7Position: ${location.x}, ${location.y}, ${location.z}`;
            const items = inv ? getItems(block, maxItemsDisplay) : undefined;
            const growth = getValidValues(block, ["growth", "culinarium:growth"]);
            if (id === "minecraft") {
                const honeyLevel = permutation?.getState("honey_level");
                const openBit = permutation?.getState("open_bit");
                const osb = permutation?.getState("output_subtract_bit");
                const repeaterDelay = permutation?.getState("repeater_delay");
                const redstone = block?.getRedstonePower();
                titleRaw += honeyLevel !== undefined ? `\n§7Honey: ${honeyLevel}/5` : "";
                titleRaw += osb !== undefined ? `\n§7Mode: ${osb ? "Subtractor" : "Comparator"}` : "";
                titleRaw += openBit !== undefined && name === "lever" ? `\n§7Active: ${openBit}` : "";
                titleRaw += repeaterDelay !== undefined ? `\n§7Delay: ${repeaterDelay + 1}` : "";
                titleRaw += redstone !== undefined && redstone > 0 ? `\n§7Power: ${redstone}` : "";
            }
            if (developer)
                underRaw = `\n§7Waterlogged: ${block.isWaterlogged}`;
            if (developer && Object.keys(states).length > 0)
                underRaw = getAllValues(states, block, underRaw);
            if (developer && block.getTags().length > 0)
                underRaw += `\n\n§7Tags: [${block.getTags().join(", ")}]`;
            player.onScreenDisplay.setActionBar({
                rawtext: [
                    { translate: specialBlocks[typeId] ?? block.localizationKey },
                    { text: displayIcon ? displayIcon : "" },
                    { text: titleRaw },
                    { translate: inv ? "waio.block.container" : "" },
                    { text: inv ? `${inv.size - inv.emptySlotsCount}/${inv.size}` : "" },
                    { translate: items?.length ? "waio.block.items" : "" },
                    ...(inv ? items : []),
                    { translate: growth ? "waio.block.growth" : "" },
                    { text: growth ? getGrowthPercentage(growth) : "" },
                    { text: underRaw },
                    { text: "\n§9§o" },
                    { translate: pack },
                ],
            });
        }
    }
});
world.beforeEvents.chatSend.subscribe((ev) => {
    const player = ev.sender;
    const message = ev.message.toLowerCase();
    if (message === "waio!configuration" || message === "w!configuration") {
        ev.cancel = true;
        system.run(async function () {
            const waio = player.getDynamicProperty("waio");
            const noLang = player.getDynamicProperty("noLang");
            const maxDistance = player.getDynamicProperty("maxDistance");
            const maxItemsDisplay = player.getDynamicProperty("maxItemsDisplay");
            const developer = player.getDynamicProperty("developer");
            const configurationForm = new ModalFormData();
            configurationForm.title("WAIO Configuration");
            configurationForm.label("What Am I Observing?");
            configurationForm.toggle({ translate: "waio.ui.active" }, { defaultValue: waio });
            configurationForm.divider();
            configurationForm.label({ translate: "waio.ui.label.personal_preferences" });
            configurationForm.toggle({ translate: "waio.ui.no_lang" }, { defaultValue: noLang, tooltip: "If disabled, the addon identifier will be formatted and used as addon name." });
            configurationForm.slider({ translate: "waio.ui.max_distance" }, 1, 15, { defaultValue: maxDistance, tooltip: "Modifies the maximum distance (in blocks) at which you will get information about what you are observing." });
            configurationForm.slider({ translate: "waio.ui.max_items_display" }, 1, 10, { defaultValue: maxItemsDisplay, tooltip: "Modifies the maximum number of items that the containers will display (if the value is higher, it is usually more uncomfortable)." });
            configurationForm.divider();
            configurationForm.label({ translate: "waio.ui.label.developer_mode" });
            configurationForm.toggle({ translate: "waio.ui.active" }, { defaultValue: developer, tooltip: "If enabled, it will allow you to see extra information, usually only useful for developers." });
            const response = await forceShow(player, configurationForm);
            if (response.canceled)
                return;
            player.setDynamicProperty("waio", response.formValues[1]);
            player.setDynamicProperty("noLang", response.formValues[4]);
            player.setDynamicProperty("maxDistance", response.formValues[5]);
            player.setDynamicProperty("maxItemsDisplay", response.formValues[6]);
            player.setDynamicProperty("developer", response.formValues[9]);
            player.sendMessage("§aConfiguration has been successfully updated.");
            player.sendMessage("§aConfiguration has been successfully updated.");
        });
    }
});
export async function forceShow(player, form, timeout = Infinity) {
    const startTick = system.currentTick;
    while (system.currentTick - startTick < timeout) {
        const response = await form.show(player);
        if (response.cancelationReason !== FormCancelationReason.UserBusy) {
            return response;
        }
    }
    throw new Error(`Timed out after ${timeout} ticks`);
}
